# -*- coding: utf-8 -*-
'''
This module contains classes for game objects that are relevant for both client and server:
*GameState*, *GameStateUpdate* and *ClientActivity*.
Client as well as server are supposed to define subclasses of classes in this module,
that extend those types with data and functionality that is client-/server-specific.

Also, this module defines the network protocol which *Connection*s and *Server*s use
to communicate. It contains the *UDPPackage* class, which represents a unit of information
that client and server can exchange with each other, as well as classes that make up parts
of a package.
'''

import sys
import random
import umsgpack

# Unique 4-byte token to mark the end of the header of a UDPPackage
_HEADER_END_TOKEN = bytes.fromhex('b5968459')
# Unique 4-byte token to mark
TO_DELETE = bytes.fromhex('d281e5ba')

class TypeClass:
    '''
    Mixin that allows to add class variables to a class during runtime.
    It is used to make enum classes like AktivityType or PackageType extensible.
    '''
    @classmethod
    def add_type(cls, name: str):
        '''
        Add a new named type to this enum-like class.
        '''
        setattr(cls, name, len(cls.__dict__) + 1)

class Sendable:
    '''
    Mixin for classes that are supposed to be sendable as part of a server request or response.
    Sendables can only have basic Python types as attributes and their constructor needs
    to be callable without passing any arguments.
    '''

    def to_bytes(self):
        '''
        Packs and return a small a binary representation of self.

        '''
        return umsgpack.packb(self.__dict__)

    @classmethod
    def from_bytes(cls, bytepack: bytes):
        '''
        Returns a copy of the object that was packed into byte format.
        '''
        try:
            received_sendable = cls()
            received_sendable.__dict__ = umsgpack.unpackb(bytepack)
            return received_sendable
        except (umsgpack.InsufficientDataException, KeyError):
            raise TypeError('Bytes could no be parsed into ' + cls.__name__ + '.')

class GameStatus(TypeClass):
    '''
    Enum class with the values:
    - *Paused*
    - *Active*
    '''
    Paused = 1
    Active = 2

class ActivityType(TypeClass):
    '''
    Enum class with the values:
    - *PauseGame*
    - *ResumeGame*

    Can and is supposed to be extended using *TypeClass*'s *add_type* method.
    '''

    PauseGame = 1
    ResumeGame = 2
    JoinServer = 3
    LeaveServer = 4

class PackageType(TypeClass):
    '''
    Enum class with the following values:
    - *GetGameStateRequest*: As client request the full shared game state from the server.
        body = None
    - *GetGameStateUpdateRequest*: As client request all polled game state updates.
        body = *GameStateUpdate* (purely for time-ordering)
    - *PostClientActivityRequest*: As client post client-side activity to the *GameService*.
        body = *ClientActivity*
    - *ServerResponse*: As server respond to a client request.
        body = request-dependent
    - *ServerError*: As server report an error to the client.
        body = *ErrorMessage*
    '''

    GetGameStateRequest = 1
    GetGameStateUpdateRequest = 2
    PostClientActivityRequest = 3
    ServerResponse = 4
    ServerError = 5

class ErrorType(TypeClass):
    '''
    Enum class with the following values:
    - *RequestTimeout*: Server _response took to long.
    - *UnpackError*: Request or _response bytepack corrupted.
    - *RequestInvalid*: Server could not handle the request.

    To be used as part of an *ErrorMessage* object in a *ServerError* package.
    '''

    RequestTimeout = 1
    UnpackError = 2
    RequestInvalid = 3
    OutOfSync = 4

class ErrorMessage(Sendable):
    '''
    The sendable type *ErrorMessage* is used for the body of *UDPPackage*s with
    *package_type* *PackageType.ServerError* in their *header*.
    '''
    def __init__(self, error_type=ErrorType.RequestInvalid, message=''):
        self.error_type = error_type
        self.message = message

class UDPPackageHeader(Sendable):
    def __init__(self, package_type=PackageType.ServerResponse, body_type='NoneType'):
        self.package_type = package_type
        self.body_type = body_type

    '''
    Override 'object' members
    '''
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.package_type == other.package_type and self.body_type == other.body_type
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

class UDPPackage:
    '''
    Contains *header* and *body* as attributes. The header contains information about the the
    package type and body. The body is some object of a core class like *ErrorMessage*,
    *GameState*, *GameStateUpdate* or *ClientActivity*.
    '''
    def __init__(self, package_type: PackageType, body: Sendable = None):
        self.header = UDPPackageHeader(package_type, body.__class__.__name__)
        self.body = body

    def to_datagram(self):
        '''
        Returns a bytepacked datagram representing the UDPPackage.
        '''
        self.header.body_type = self.body.__class__.__name__
        datagram = bytearray(self.header.to_bytes())
        datagram.extend(_HEADER_END_TOKEN)
        if self.body != None:
            datagram.extend(self.body.to_bytes())
        return bytes(datagram)

    @staticmethod
    def from_datagram(datagram: bytes):
        '''
        Unpacks the given bytepacked datagram and returns it's content as a *UDPPackage*
        object.
        '''
        datagram = datagram.split(_HEADER_END_TOKEN)
        header = UDPPackageHeader.from_bytes(datagram[0])
        if header.body_type != 'NoneType':
            body = getattr(sys.modules[__name__], header.body_type).from_bytes(datagram[1])
        else:
            body = None
        return UDPPackage(header.package_type, body)

    def is_response(self):
        '''
        Returns *True* if the package is of package type *ServerResponse*.
        '''
        return self.header.package_type == PackageType.ServerResponse

    def is_error(self):
        '''
        Returns *True* if the package is of package type *ServerError*.
        '''
        return self.header.package_type == PackageType.ServerError

    def is_update_request(self):
        '''
        Returns *True* if the package is of package type *GetGameStateUpdateRequest*.
        '''
        return self.header.package_type == PackageType.GetGameStateUpdateRequest

    def is_state_request(self):
        '''
        Returns *True* if the package is of package type *GetGameStateRequest*.
        '''
        return self.header.package_type == PackageType.GetGameStateRequest

    def is_post_activity_request(self):
        '''
        Returns *True* if the package is of package type *PostClientActivityRequest*.
        '''
        return self.header.package_type == PackageType.PostClientActivityRequest

    '''
    Override object members
    '''
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.header == other.header and self.body == other.body
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

class GameState(Sendable):
    '''
    Contains game state information that is required to be known both by the server and the client.
    Since it is a *Sendable*, it can only contain basic python types as attributes.

    *time_order* should be in alignment with the servers current update counter.
    '''

    def __init__(self, time_order=0, game_status=GameStatus.Paused):
        self.game_status = game_status
        self.time_order = time_order
        self.players = {}

        ### ONLY FOR TESTING PURPOSES
        self.test_pos = 0

    def is_paused(self):
        '''
        Returns *True* if game status is *Paused*.
        '''
        return self.game_status == GameStatus.Paused

    def iter(self, dict_attr: str):
        '''
        Returns a representation of the *GameState* attribute *dict_attr* that is safe to
        iterate through, while the GameState is concurrently updated. Returns keys and
        values in a tuple.

        Example: Iterate through players with `for player_id, player in game_state.iter('players')`.
        '''
        return getattr(self, dict_attr).copy().items()

    '''
    Overrides of 'object' member functions
    '''
    # Check time ordering
    def __lt__(self, other):
        return self.time_order < other.time_order

    def __gt__(self, other):
        return self.time_order > other.time_order

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.game_status == other.game_status
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

class GameStateUpdate(Sendable):
    '''
    Represents a set of changes to carry out on a *GameState*.
    The server keep a *time_order* counter and labels all updates in ascending order.

    Keywords are *GameState* atttribute names. If you want to remove some key from the
    game state (*GameState* attributes themselves can not be deleted), just assign
    *TO_DELETE* to it in the update.

    Use the *+* operator to add *GameStateUpdate*s together or to add them to a
    *GameState* (returning the updated update/state).

    Adding up available updates will always result in an equally or more current but
    also heavier update (meaning it will contain more data).
    '''

    def __init__(self, time_order=0, **kwargs):
        self.__dict__ = kwargs
        self.time_order = time_order

    # Adding to another update should return an updated update
    def __add__(self, other):
        if other > self:
            _recursive_update(self.__dict__, other.__dict__)
            return self
        else:
            _recursive_update(other.__dict__, self.__dict__)
            return other

    # Adding to a GameState should update and return the state
    def __radd__(self, other):
        if type(other) is int:
            # This is so that sum() works on lists of updates, because
            # sum will always begin by adding the first element of the
            # list to 0.
            return self
        if self > other:
            _recursive_update(other.__dict__, self.__dict__, delete=True)
        return other

    # Check time ordering
    def __lt__(self, other):
        return self.time_order < other.time_order

    def __gt__(self, other):
        return self.time_order > other.time_order

class ClientActivity(Sendable):
    '''
    An update the client sends to the server about client-side processes like player movement and
    collisions. The server will validate *ClientActivity* samples and respond with an *OutOfSync*
    error, if they contradict the server-side game state.

    *activity_data* is a *dict* object, that contains all necessary information to process the
    activity server-side (a player's *id*, *position* and *velocity* for example).
    '''
    def __init__(self, activity_type=ActivityType.PauseGame, activity_data={}):
        self.activity_type = activity_type
        self.activity_data = activity_data

def join_server_activity(player_name: str):
    '''
    Returns a *ClientActivity* that joins a player with name *player_name* to the game.

    The field *join_id* in the *activity_data* attribute of the activity can be used
    to uniquely identify the joined player in the shared game state. This means player
    names need not be unique.
    '''
    join_id = ''.join(random.choices('0123456789abcdef', k=8))
    return ClientActivity(
        activity_type=ActivityType.JoinServer,
        activity_data={
            'name': player_name,
            'join_id': join_id
        }
    )

def leave_server_activity(player_id):
    '''
    Returns a *ClientActivity* that removes the player with the ID *player_id* from the game.
    '''
    return ClientActivity(
        activity_type=ActivityType.LeaveServer,
        activity_data={'player_id': player_id}
    )

def timeout_error(message=''):
    '''
    Returns a *UDPPackage* with package type *ServerError*,
    error type *RequestTimeout* and *message* as error message.
    '''
    return UDPPackage(
        package_type=PackageType.ServerError,
        body=ErrorMessage(
            error_type=ErrorType.RequestTimeout,
            message=message
        )
    )

def unpack_error(message=''):
    '''
    Returns a *UDPPackage* with package type *ServerError*,
    error type *UnpackError* and *message* as error message.
    '''
    return UDPPackage(
        package_type=PackageType.ServerError,
        body=ErrorMessage(
            error_type=ErrorType.UnpackError,
            message=message
        )
    )

def request_invalid_error(message=''):
    '''
    Returns a *UDPPackage* with package type *ServerError*,
    error type *RequestInvalid* and *message* as error message.
    '''
    return UDPPackage(
        package_type=PackageType.ServerError,
        body=ErrorMessage(
            error_type=ErrorType.RequestInvalid,
            message=message
        )
    )

def out_of_sync_error(message=''):
    '''
    Returns a *UDPPackage* with package type *ServerError*,
    error type *OutOfSync* and *message* as error message.
    '''
    return UDPPackage(
        package_type=PackageType.ServerError,
        body=ErrorMessage(
            error_type=ErrorType.OutOfSync,
            message=message
        )
    )

def game_state_request():
    '''
    Returns a *UDPPackage* with package type *GetGameStateRequest*.
    '''
    return UDPPackage(
        package_type=PackageType.GetGameStateRequest
    )

def game_state_update_request(time_order: int):
    '''
    Returns a *UDPPackage* with package type *GetGameStateUpdateRequest*.
    Enter the *time_order* attribute of the client's last known *GameState*.
    '''
    return UDPPackage(
        package_type=PackageType.GetGameStateUpdateRequest,
        body=GameStateUpdate(
            time_order=time_order
        )
    )

def post_activity_request(client_activity: ClientActivity):
    '''
    Returns a *UDPPackage* with package type *PostClientActivityRequest* with
    the given *ClientActivity* object as it's body.
    '''
    return UDPPackage(
        package_type=PackageType.PostClientActivityRequest,
        body=client_activity
    )

def response(body: Sendable):
    '''
    Returns a *UDPPackage* with package type *ServerResponse*.
    '''
    return UDPPackage(
        package_type=PackageType.ServerResponse,
        body=body
    )

def toggle_pause_activity(shared_game_state: GameState):
    '''
    Returns a *ClientActivity* that either pauses or resumes the server's game loop, depending
    on the *game_status* of the given *GameState*.
    '''
    if shared_game_state.is_paused():
        activity_type = ActivityType.ResumeGame
    else:
        activity_type = ActivityType.PauseGame
    return ClientActivity(
        activity_type=activity_type,
        activity_data={}
    )

# This is for GameStateUpdate objects, which should update nested
# dicts recursively so that no state data is unexpectedly deleted.
# Also, this functions manages deletion of state objects.
def _recursive_update(d: dict, u: dict, delete=False):
    for k, v in u.items():
        if v == TO_DELETE and delete:
            del d[k]
        elif isinstance(v, dict):
            d[k] = _recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d



# -*- coding: utf-8 -*-
'''
This module mainly contains the *Connection* class, which represents a connection to a
running Server (meaning a game server). Use this to manage your server connections.

**Note: If you want to connect to a server in another local network you must use the proper IPv4
address of that network, and not the local IP address of the server. Also the port on port on
which the *Server* serves has to be properly forwarded within that network.**
'''

import socket
import threading
import time
import pygase.shared

REQUEST_TIMEOUT = 0.5
CONNECTION_TIMEOUT = 5.0
_BUFFER_SIZE = 1024

class ConnectionStatus(pygase.shared.TypeClass):
    '''
    Enum class with the following values:
    - *Connected*: Connection is running.
    - *WaitingForServer*: Connection is trying to connect/reconnect to the server.
    - *Disconnected*: Connection is not communicating with the server.
    '''
    
    Connected = 1
    WaitingForServer = 2
    Disconnected = 3

class Connection:
    '''
    Initialization of a *Connection* will open a connection to a BossFight Server
    with the specified *server_address* as a tuple containing the IP-adress as a string and the
    port as an int. Check the *connection_status* attribute to get the status of the Connection as
    a *ConnectionStatus* attribute.

    A running *Connection* will request an update of *game_state* from the server
    every *update_cycle_interval* seconds.
    '''
    def __init__(self, server_address, closed=False):
        self.game_state = pygase.shared.GameState()
        self.server_address = server_address
        self._client_socket = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)
        self._client_socket.settimeout(REQUEST_TIMEOUT)
        self.latency = 0
        self._polled_client_activities = []
        self.update_cycle_interval = 0.02
        if not closed:
            self._update_cycle_thread = threading.Thread(target=self._update_cycle)
            self.connection_status = ConnectionStatus.WaitingForServer
            self._update_cycle_thread.start()
        else:
            self._update_cycle_thread = threading.Thread()
            self.connection_status = ConnectionStatus.Disconnected

    def __del__(self):
        self.disconnect()
        self._client_socket.close()

    def _send_and_recv(self, package: pygase.shared.UDPPackage):
        # Send package to server ...
        self._client_socket.sendto(package.to_datagram(), self.server_address)
        # ... and get response if possible, otherwise create ServerError package
        try:
            return pygase.shared.UDPPackage.from_datagram(self._client_socket.recv(_BUFFER_SIZE))
        except socket.timeout:
            return pygase.shared.timeout_error('Request timed out.')
        except ConnectionResetError:
            return pygase.shared.timeout_error('Server not found.')

    def connect(self):
        '''
        Will try to connect/reconnect to the server if *connection_status* is
        *ConnectionStatus.Disconnected*, otherwise does nothing.
        '''
        if not self._update_cycle_thread.is_alive():
            self._update_cycle_thread = threading.Thread(target=self._update_cycle)
            self._update_cycle_thread.start()

    def disconnect(self):
        '''
        Will stop the connection from sending any further requests to the server.
        Will do nothing if *connection_status* == *ConnectionStatus.Disconnected*.
        '''
        t_0 = time.time()
        # Force update cycle to end
        while self._update_cycle_thread.is_alive() and time.time()-t_0 < REQUEST_TIMEOUT:
            self.connection_status = ConnectionStatus.Disconnected
        if self._update_cycle_thread.is_alive():
            raise threading.ThreadError

    def is_connected(self):
        '''
        Returns *True* if the connection status is *Connected*.
        '''
        return self.connection_status == ConnectionStatus.Connected

    def is_waiting(self):
        '''
        Returns *True* if the connection status is *WaitingForServer*.
        '''
        return self.connection_status == ConnectionStatus.WaitingForServer

    def post_client_activity(self, client_activity: pygase.shared.ClientActivity):
        '''
        Sends the *ClientActivity* object to the server.
        '''
        self._polled_client_activities.append(client_activity)

    def _try_connect(self):
        t_0 = time.time()
        self.connection_status = ConnectionStatus.WaitingForServer
        # Try to successfully get the shared game state for connection_timeout seconds
        while time.time()-t_0 < CONNECTION_TIMEOUT and \
          not self.connection_status == ConnectionStatus.Disconnected:
            t_1 = time.time()
            response = self._send_and_recv(pygase.shared.game_state_request())
            if response.is_response():
                self.game_state = response.body
                if not self.connection_status == ConnectionStatus.Disconnected:
                    self.connection_status = ConnectionStatus.Connected
                return # Connection successful, leave _try_connect()
            #elif response.is_error():
            #    print(response.body.message)
            dt = time.time() - t_1
            time.sleep(max(self.update_cycle_interval - dt, 0))
        # if this point is reached connection was unsuccessful
        self.connection_status = ConnectionStatus.Disconnected

    def _update_cycle(self):
        self._try_connect()
        latency_timer = 0
        while not self.connection_status == ConnectionStatus.Disconnected:
            t_0 = time.time()
            # Post activities first
            activities_to_post = self._polled_client_activities[:5] # First 5 activities in queue
            for activity in activities_to_post:
                response = self._send_and_recv(
                    pygase.shared.post_activity_request(activity)
                )
                if response.is_response():
                    self._polled_client_activities.remove(activity)
            # Then get game state update
            response = self._send_and_recv(
                pygase.shared.game_state_update_request(self.game_state.time_order)
            )
            if response.is_response():
                self.game_state += response.body
            else:
                #if response.is_error():
                #    print(response.body.message)
                self._try_connect()
            dt = time.time()-t_0
            if not latency_timer % 60:
                self.latency = dt*1000
            latency_timer += 1
            time.sleep(max(self.update_cycle_interval-dt, 0))



# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='pygase',
    version='0.1.7',
    description='A lightweight client-server technology and UDP-based network protocol for real-time online gaming.',
    author='Silas Bischoff',
    author_email='silas.bischoff@stud.uni-due.de',
    license='MIT',
    url='https://github.com/sbischoff-ai/python-game-service',
    keywords=['server', 'client', 'games', 'gaming', 'real-time', 'network', 'UDP', 'protocol', 'game server', 'game service', 'gameservice'],
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Topic :: Games/Entertainment',
        'Topic :: Internet',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
        'Intended Audience :: Developers',
        'Development Status :: 3 - Alpha'
    ],
    packages=['pygase'],
    install_requires=['u-msgpack-python']
)


# -*- coding: utf-8 -*-
'''
This module defines the *Server* class, a server that can handle requests from
a client's *Connections* object and the *GameLoop*, which simulates the game logic.

**Note: The IP address you bind the Server to is a local IP address from the
192.168.x.x address space. If you want computers outside your local network to be
able to connect to your game server, you will have to forward the port from the local
address your server is bound to to your external IPv4 address!**
'''

import socketserver
import threading
import time
import pygase.shared

UPDATE_CACHE_SIZE = 100

class Server(socketserver.ThreadingUDPServer):
    '''
    Threading UDP server that manages clients and processes requests.
    *game_loop_class* is your subclass of *GameLoop*, which implements the handling of activities
    and the game state update function. *game_state* is an instance of *pygase.shared.GameState*
    that holds all necessary initial data.

    Call *start()* for the server to start handling requests from
    *Connections*s. Call *shutdown()* to stop it.

    *game_loop* is the server's *GameLoop* object, which simulates the game logic and updates
    the *game_state*.
    '''

    def __init__(self, ip_address: str, port: int, game_loop_class: type, game_state: pygase.shared.GameState):
        super().__init__((ip_address, port), ServerRequestHandler)
        self.game_state = game_state
        self._client_activity_queue = []
        self._state_update_cache = [pygase.shared.GameStateUpdate()]
        self._join_counter = 0
        self.game_loop = game_loop_class(self)
        self._server_thread = threading.Thread()

    def start(self):
        '''
        Runs the server in a dedicated Thread and starts the game loop.
        Does nothing if server is already running.
        Must be called for the server to handle requests and is terminated by *shutdown()*
        '''
        if not self._server_thread.is_alive():
            self._server_thread = threading.Thread(target=self.serve_forever)
            self.game_loop.start()
            self._server_thread.start()

    def shutdown(self):
        '''
        Stops the server's request handling and pauses the game loop.
        '''
        super().shutdown()
        self.game_loop.pause()

    def get_ip_address(self):
        '''
        Returns the servers IP address as a string.
        '''
        return self.socket.getsockname()[0]

    def get_port(self):
        '''
        Returns the servers port as an integer.
        '''
        return self.socket.getsockname()[1]

class ServerRequestHandler(socketserver.BaseRequestHandler):
    def handle(self):
        # Read out request
        try:
            request = pygase.shared.UDPPackage.from_datagram(self.request[0])
        except TypeError:
            # If unpacking request failed send back error message and exit handle function
            response = pygase.shared.unpack_error('Server responded: Byte error.')
            self.request[1].sendto(response.to_datagram(), self.client_address)
            return

        # Handle request and assign response here
        if request.is_update_request():
            if request.body < min(self.server._state_update_cache):
                # If the client's time_order is too old, respond with error so the client resyncs.
                response = pygase.shared.out_of_sync_error('Server responded: Lag error.')
                self.request[1].sendto(response.to_datagram(), self.client_address)
                return
            # Respond by sending the sum of all updates since the client's time-order point.
            update = sum(
                (upd for upd in self.server._state_update_cache if upd > request.body),
                request.body
            )
            response = pygase.shared.response(update)
        elif request.is_post_activity_request():
            # Pausing or Resuming the game and joining a server must be possible outside
            # of the game loop:
            if request.body.activity_type == pygase.shared.ActivityType.PauseGame:
                self.server.game_loop.pause()
                self.server.game_state.time_order += 1
                self.server.game_loop._cache_state_update(pygase.shared.GameStateUpdate(
                    time_order=self.server.game_state.time_order,
                    game_status=pygase.shared.GameStatus.Paused
                ))
            elif request.body.activity_type == pygase.shared.ActivityType.ResumeGame:
                self.server.game_state.time_order += 1
                self.server.game_loop._cache_state_update(pygase.shared.GameStateUpdate(
                    time_order=self.server.game_state.time_order,
                    game_status=pygase.shared.GameStatus.Active
                ))
                self.server.game_loop.start()
            elif request.body.activity_type == pygase.shared.ActivityType.JoinServer \
              and self.server.game_state.is_paused():
                # A player dict is added to the game state. The id is unique for the
                # server session (counting from 0 upwards).
                update = pygase.shared.GameStateUpdate(
                    self.server.game_state.time_order + 1
                )
                self.server.game_loop._add_player(request.body, update)
                self.server.game_state += update
                self.server.game_loop._cache_state_update(update)
            elif request.body.activity_type == pygase.shared.ActivityType.LeaveServer \
              and self.server.game_state.is_paused():
                update = pygase.shared.GameStateUpdate(self.server.game_state.time_order + 1)
                update.players = {
                    request.body.activity_data['player_id']: pygase.shared.TO_DELETE
                }
                self.server.game_state += update
                self.server.game_loop._cache_state_update(update)
            else:
                # Any other kind of activity: add to the queue for the game loop
                self.server._client_activity_queue.append(request.body)
            response = pygase.shared.response(None)
        elif request.is_state_request():
            # respond by sending back the shared game state
            response = pygase.shared.response(self.server.game_state)
        else:
            # if none of the above were a match the request was invalid
            response = pygase.shared.request_invalid_error('Server responded: Request invalid.')

        # Send response
        self.request[1].sendto(response.to_datagram(), self.client_address)

class GameLoop:
    '''
    Class that can update a shared game state by running a game logic simulation thread.

    You should inherit from this class and implement the *handle_activity()* and
    *update_game_state()* methods, as well as *on_join()*.
    '''
    def __init__(self, server: Server):
        self.server = server
        self.game_time = 0.0
        self._game_loop_thread = threading.Thread()
        self.update_cycle_interval = 0.02

    def start(self):
        '''
        Starts a thread that updates the shared game state every *update_cycle_interval* seconds.
        Use this to restart a paused game.
        '''
        if not self._game_loop_thread.is_alive():
            self._game_loop_thread = threading.Thread(target=self._update_cycle)
            self.server.game_state.game_status = pygase.shared.GameStatus.Active
            self._game_loop_thread.start()

    def pause(self):
        '''
        Stops the game loop until *start()* is called.
        If the game loop is not currently running does nothing.
        '''
        self.server.game_state.game_status = pygase.shared.GameStatus.Paused

    def _update_cycle(self):
        dt = self.update_cycle_interval
        while not self.server.game_state.is_paused():
            t = time.time()
            # Create update object and fill it with all necessary changes
            update_by_clients = pygase.shared.GameStateUpdate(self.server.game_state.time_order + 1)
            # Handle client activities first
            activities_to_handle = self.server._client_activity_queue[:5]
            # Get first 5 activitys in queue
            for activity in activities_to_handle:
                if activity.activity_type == pygase.shared.ActivityType.JoinServer:
                    self._add_player(activity, update_by_clients)
                elif activity.activity_type == pygase.shared.ActivityType.LeaveServer:
                    player_id = activity.activity_data['player_id']
                    if hasattr(update_by_clients, 'players'):
                        update_by_clients.players[player_id] = pygase.shared.TO_DELETE
                    else:
                        update_by_clients.players = {player_id: pygase.shared.TO_DELETE}
                else:
                    self.handle_activity(activity, update_by_clients, dt)
                self.server._client_activity_queue.remove(activity)
            # Add activity update to state, cache it and then run the server state update
            self.server.game_state += update_by_clients
            self._cache_state_update(update_by_clients)
            update_by_server = pygase.shared.GameStateUpdate(self.server.game_state.time_order + 1)
            self.update_game_state(update_by_server, dt)
            # Add the final update to the game state and cache it for the clients
            self.server.game_state += update_by_server
            self._cache_state_update(update_by_server)
            dt = time.time() - t
            if dt < self.update_cycle_interval:
                time.sleep(self.update_cycle_interval-dt)
                dt = self.update_cycle_interval
            self.game_time += dt

    def _add_player(self, join_activity, update):
        player = {
            'name': join_activity.activity_data['name'],
            'join_id': join_activity.activity_data['join_id']
        }
        if hasattr(update, 'players'):
            update.players[self.server._join_counter] = player
        else:
            update.players = {self.server._join_counter: player}
        self.on_join(
            player_id=self.server._join_counter,
            update=update
        )
        self.server._join_counter += 1

    def on_join(self, player_id: int, update: pygase.shared.GameStateUpdate):
        '''
        Override this method to define your initial player data.
        *update.players[player_id]* will already be accessible when *on_join* is called.
        '''
        pass

    def handle_activity(self, activity: pygase.shared.ClientActivity, update: pygase.shared.GameStateUpdate, dt):
        '''
        Override this method to implement handling of client activities. Any state changes should be
        written into the update argument of this method.
        '''
        pass
        #raise NotImplementedError

    def update_game_state(self, update: pygase.shared.GameStateUpdate, dt):
        '''
        Override to implement an iteration of your game logic simulation.
        State changes should be written into the update argument.
        Attributes of the shared game state that do not change at all, should not
        be assigned to *update* (in order to optimize network performance).
        '''
        pass
        #raise NotImplementedError

    def _cache_state_update(self, state_update: pygase.shared.GameStateUpdate):
        self.server._state_update_cache.append(state_update)
        if len(self.server._state_update_cache) > UPDATE_CACHE_SIZE:
            self.server._state_update_cache = self.server._state_update_cache[1:]
